CREATE DATABASE Opiskelijat;
Use Opiskelijat;

DROP TABLE Tenttii
;
DROP TABLE Kurssi
;
DROP TABLE Puhelinnumero
;
DROP TABLE Opiskelija
;
DROP TABLE Laitos
;
DROP TABLE Tiedekunta
;
DROP TABLE Postinumero
;

CREATE TABLE Postinumero (
Postinumero      CHAR(5)     NOT NULL,
Postitoimipaikka VARCHAR(64) NOT NULL,
CONSTRAINT Postinumero_PK
   PRIMARY KEY (Postinumero)
)
;
CREATE TABLE Tiedekunta (
TiedekuntaID      INTEGER     NOT NULL,
Nimi              VARCHAR(64) NOT NULL,
CONSTRAINT Tiedekunta_PK
   PRIMARY KEY (TiedekuntaID)
)
;
CREATE TABLE Laitos (
LaitosID          INTEGER     NOT NULL,
Nimi              VARCHAR(64) NOT NULL,
Tiedekunta        INTEGER     NOT NULL,
CONSTRAINT Laitos_PK
   PRIMARY KEY (LaitosID),
CONSTRAINT Laitos_FK_T
   FOREIGN KEY (Tiedekunta)
   REFERENCES Tiedekunta (TiedekuntaID)
      ON UPDATE CASCADE
      ON DELETE NO ACTION
)
;
CREATE TABLE Opiskelija (
Sotu                 CHAR(11)     NOT NULL,
Sahkopostiosoite     VARCHAR(64),
Etunimi              VARCHAR(32)  NOT NULL,
Sukunimi             VARCHAR(64)  NOT NULL,
Lahiosoite           VARCHAR(64)  NOT NULL,
Postinumero          CHAR(5)      NOT NULL,
Aloitusvuosi         SMALLINT     NOT NULL,
Laitos               INTEGER      NOT NULL,
CONSTRAINT Opiskelija_PK
   PRIMARY KEY (Sotu),
CONSTRAINT Opiskelija_FK_L
   FOREIGN KEY (Laitos)
   REFERENCES Laitos (LaitosID)
      ON UPDATE CASCADE
      ON DELETE NO ACTION,
CONSTRAINT Opiskelija_FK_P
   FOREIGN KEY (Postinumero)
   REFERENCES Postinumero (Postinumero)
      ON UPDATE CASCADE
      ON DELETE NO ACTION,
CONSTRAINT Opiskelija_Check_A
   CHECK (Opiskelija.Aloitusvuosi >= 1990)
)
;
CREATE TABLE Puhelinnumero (
Sotu             CHAR(11)    NOT NULL,
Puhelinnumero    VARCHAR(32) NOT NULL,
CONSTRAINT Puhelinnumero_PK
   PRIMARY KEY (Sotu, Puhelinnumero),
CONSTRAINT Puhelinnumero_FK_L
   FOREIGN KEY (Sotu)
   REFERENCES Opiskelija (Sotu)
      ON UPDATE CASCADE
      ON DELETE CASCADE

)
;
CREATE TABLE Kurssi (
Kurssikoodi    CHAR(6)              NOT NULL,
Nimi           VARCHAR(64)          NOT NULL,
Laajuus        DOUBLE PRECISION     NOT NULL DEFAULT 1,
Laitos         INTEGER              NOT NULL,
CONSTRAINT Kurssi_PK
   PRIMARY KEY (Kurssikoodi),
CONSTRAINT Kurssi_FK_L
   FOREIGN KEY (Laitos)
   REFERENCES Laitos (LaitosID)
      ON UPDATE CASCADE
      ON DELETE NO ACTION
)
;
CREATE TABLE Tenttii (
Opiskelija   CHAR(11)           NOT NULL,
Kurssi       CHAR(6)            NOT NULL,
Paivamaara   DATE               NOT NULL, 
Arvosana     DOUBLE PRECISION   NOT NULL DEFAULT 1,
CONSTRAINT Tenttii_PK
   PRIMARY KEY (Opiskelija, Kurssi, Paivamaara),
CONSTRAINT Tenttii_FK_O
   FOREIGN KEY (Opiskelija)
   REFERENCES Opiskelija (Sotu)
      ON UPDATE CASCADE
      ON DELETE NO ACTION,
CONSTRAINT Tenttii_FK_K
   FOREIGN KEY (Kurssi)
   REFERENCES Kurssi (Kurssikoodi)
      ON UPDATE CASCADE
      ON DELETE NO ACTION
);


--Sisältö

DELETE FROM Tenttii
;
DELETE FROM Kurssi
;
DELETE FROM Puhelinnumero
;
DELETE FROM Opiskelija
;
DELETE FROM Postinumero
;
DELETE FROM Laitos
;
DELETE FROM Tiedekunta
;

INSERT INTO Tiedekunta VALUES (1, 'Informaatioteknologian tiedekunta');
INSERT INTO Tiedekunta VALUES (2, 'Humanistinen tiedekunta');
INSERT INTO Tiedekunta VALUES (3, 'Kasvatustieteiden tiedekunta');

INSERT INTO Laitos VALUES (1, 'Tietotekniikan laitos', 1);
INSERT INTO Laitos VALUES (2, 'Tietojarjestelmatieteen laitos', 1);
INSERT INTO Laitos VALUES (3, 'Englannin kielen laitos', 2);
INSERT INTO Laitos VALUES (4, 'Kirjallisuuden laitos', 2);
INSERT INTO Laitos VALUES (5, 'Kasvatustieteen laitos', 3);
INSERT INTO Laitos VALUES (6, 'Opettajankoulutuslaitos', 3);

INSERT INTO Postinumero VALUES ('40700', 'Kajaani');
INSERT INTO Postinumero VALUES ('40740', 'Kajaani');
INSERT INTO Postinumero VALUES ('40351', 'Kajaani');
INSERT INTO Postinumero VALUES ('31600', 'Jokioinen');
INSERT INTO Postinumero VALUES ('74700', 'Kiuruvesi');

INSERT INTO Opiskelija VALUES ('111111-111P', 'eero.huusko@kamk.fi', 'Eero', 'Huusko', 'TNT 9 a 666', '40740', 1995, 1);
INSERT INTO Opiskelija VALUES ('121212-112P', 'markkuk@kamk.fi', 'Markku', 'Karppinen', 'Agora', '40351', 1992, 1);
INSERT INTO Opiskelija VALUES ('230173-000L', 'foo@bar.invalid', 'Teemu', 'Tieteilija', 'TNT 9 Y 123', '40740', 2016, 2);
INSERT INTO Opiskelija VALUES ('010101-011P', 'bar@foo.invalid', 'Taina', 'Tieteilija', 'TNT 9 Y 124', '40740', 1996, 2);
INSERT INTO Opiskelija VALUES ('111112-111P', 'barfoo@bar.invalid', 'Kaisa', 'Kielinero', 'Keskusta', '40700', 1995, 3);
INSERT INTO Opiskelija VALUES ('111113-111P', 'barfoo@bar.invalid', 'Keijo', 'Kirjailija', 'Keskusta', '40700', 2017, 4);
INSERT INTO Opiskelija VALUES ('111114-111P', NULL, 'Kimmo', 'Kasvattaja', 'TNT 9 A 000', '40740', 2018, 5);
INSERT INTO Opiskelija VALUES ('111115-111P', NULL, 'Olli', 'Opettaja', 'Keskusta', '40700', 2019, 6);
INSERT INTO Opiskelija VALUES ('111116-1116', NULL, 'Data', 'Analyytikko', 'Perukka', '40700', 2019, 6);
INSERT INTO Opiskelija VALUES ('111116-1117', NULL, 'Data', 'Analyytikko', 'Perukka', '40700', 2019, 6);


INSERT INTO Puhelinnumero VALUES ('111111-111P', '050-1232746');
INSERT INTO Puhelinnumero VALUES ('121212-112P', '050-5858585');
INSERT INTO Puhelinnumero VALUES ('230173-000L', '040-123454');
INSERT INTO Puhelinnumero VALUES ('111112-111P', '040-986238');
INSERT INTO Puhelinnumero VALUES ('230173-000L', '040-891768');
INSERT INTO Puhelinnumero VALUES ('111115-111P', '050-278823');
INSERT INTO Puhelinnumero VALUES ('111116-1116', '050-256889');
INSERT INTO Puhelinnumero VALUES ('111116-1117', NULL);

INSERT INTO Kurssi VALUES ('TIE150', 'Tietokannat', 2, 1);
INSERT INTO Kurssi VALUES ('TIE160', 'Mikrotietokoneiden ohjelmistot', 2, 1);
INSERT INTO Kurssi VALUES ('TIE001', 'Tietotekniikan propedeuttinen', 2, 1);
INSERT INTO Kurssi VALUES ('TIE101', 'Tietoverkot työvalineenä', 2, 1);
INSERT INTO Kurssi VALUES ('TIE140', 'Mikrotietokonelaitteistot', 2, 1);
INSERT INTO Kurssi VALUES ('TIE110', 'Johdatus ohjelmointiin', 3, 1);
INSERT INTO Kurssi VALUES ('TIE120', 'Ohjelmointi++', 5, 1);
INSERT INTO Kurssi VALUES ('TIE130', 'Tietotekniikan perusteet', 3, 1);

INSERT INTO Tenttii VALUES ('111111-111P', 'TIE150', '2017-01-01', 3);
INSERT INTO Tenttii VALUES ('121212-112P', 'TIE150', '2018-01-01', 2.75);
INSERT INTO Tenttii VALUES ('230173-000L', 'TIE150', '2016-01-01', 2.5);
INSERT INTO Tenttii VALUES ('010101-011P', 'TIE150', '2017-01-01', 2.25);
INSERT INTO Tenttii VALUES ('111112-111P', 'TIE150', '2017-01-01', 2);
INSERT INTO Tenttii VALUES ('111112-111P', 'TIE150', '2017-03-01', 2.75);
INSERT INTO Tenttii VALUES ('111113-111P', 'TIE150', '2017-01-01', 1.75);
INSERT INTO Tenttii VALUES ('111114-111P', 'TIE150', '2017-01-01', 1.5);
INSERT INTO Tenttii VALUES ('111114-111P', 'TIE150', '2017-02-01', 2.25);
INSERT INTO Tenttii VALUES ('111111-111P', 'TIE160', '2019-01-01', 2);
INSERT INTO Tenttii VALUES ('121212-112P', 'TIE160', '2019-01-01', 1.75);
INSERT INTO Tenttii VALUES ('230173-000L', 'TIE160', '2019-01-01', 1.5);
INSERT INTO Tenttii VALUES ('010101-011P', 'TIE160', '2019-01-01', 1.25);
INSERT INTO Tenttii VALUES ('111112-111P', 'TIE160', '2019-01-01', 3);
INSERT INTO Tenttii VALUES ('111113-111P', 'TIE160', '2019-01-01', 2.75);
INSERT INTO Tenttii VALUES ('111114-111P', 'TIE160', '2019-01-01', 2.5);
INSERT INTO Tenttii VALUES ('111111-111P', 'TIE110', '2019-01-01', 2);
INSERT INTO Tenttii VALUES ('121212-112P', 'TIE110', '2019-02-01', 0);
INSERT INTO Tenttii VALUES ('230173-000L', 'TIE110', '2019-03-01', 2);
INSERT INTO Tenttii VALUES ('010101-011P', 'TIE110', '2019-04-01', 0);
INSERT INTO Tenttii VALUES ('111112-111P', 'TIE110', '2019-05-01', 3);
INSERT INTO Tenttii VALUES ('111113-111P', 'TIE110', '2019-06-01', 0);
INSERT INTO Tenttii VALUES ('111114-111P', 'TIE110', '2019-07-01', 0);
INSERT INTO Tenttii VALUES ('111111-111P', 'TIE101', '2019-01-01', 0);
INSERT INTO Tenttii VALUES ('121212-112P', 'TIE101', '2019-01-01', 1.75);
INSERT INTO Tenttii VALUES ('230173-000L', 'TIE001', '2019-01-01', 1.5);
INSERT INTO Tenttii VALUES ('010101-011P', 'TIE001', '2019-01-01', 1.25);
INSERT INTO Tenttii VALUES ('111112-111P', 'TIE001', '2019-01-01', 0);
INSERT INTO Tenttii VALUES ('111113-111P', 'TIE120', '2019-01-01', 0);
INSERT INTO Tenttii VALUES ('111114-111P', 'TIE120', '2019-01-01', 2.5);
INSERT INTO Tenttii VALUES ('111111-111P', 'TIE101', '2019-02-01', 1);
INSERT INTO Tenttii VALUES ('121212-112P', 'TIE101', '2019-02-01', 2.75);
INSERT INTO Tenttii VALUES ('230173-000L', 'TIE001', '2019-02-01', 2.5);
INSERT INTO Tenttii VALUES ('010101-011P', 'TIE001', '2019-02-01', 2.25);
INSERT INTO Tenttii VALUES ('111112-111P', 'TIE001', '2019-02-01', 2);
INSERT INTO Tenttii VALUES ('111113-111P', 'TIE120', '2019-02-01', 3);
INSERT INTO Tenttii VALUES ('111114-111P', 'TIE120', '2019-02-01', 2.75);





