# 08: Transaktiot ja Näkymät

![alt text](images/08_transaktiot_ja_nakymat_header.jpg "Transaktiot ja näkymät")

Sisällysluettelo:

[[_TOC_]]

Linkit-osiossa linkkejä ulkoisiin lähteisiin, joissa käydään tunnin asioita läpi englanniksi.

# MySQL - Osa 6 - Indeksien aloitus ja Transaktiot

Sisältö:
- Indeksoinnin periaate
- Tapahtumien hallinta (Transaktiot)

## Vähän teoriaa alkuun
ACID (engl. Atomicity, Consistency, Isolation, Durability, suom. Atomisuus, eheys, eristyneisyys, pysyvyys) tarkoittaa tietokantajärjestelmien periaatetta, jonka avulla turvataan järjestelmän tietojen eheys kaikissa tilanteissa.

* Atomisuus tarkoittaa, että transaktio suoritetaan joko kokonaan tai ei lainkaan.
* Eheys takaa, että tietokanta transaktioiden myötä siirtyy yhdestä eheästä tilasta toiseen eheään tilaan.
* Eristyneisyys takaa, että transaktiot eivät vaikuta toisiinsa ja toimivat kuin yksin järjestelmässä.
* Pysyvyys määrää, että transaktion jälkeen muutokset eivät enää voi kadota järjestelmästä

## Mysql - Indeksit 

Indeksit ovat eräänlaisia hakemistoja sarakkeille. SQL:ssä voidaan tietoja hakea ja järjestää minkä tahansa sarakkeiden mukaan. Käytettäessä tietokantaa indeksien käyttö nopeuttaa hakuja, jos ne määritellään koskemaan usein hakuehtoina käytettyjä kenttiä, tai kenttiä joiden mukaan hakutuloksia järjestetään. Indeksejä voi perustaa tai poistaa milloin tahansa ilman, että tietokantaa käyttäviä ohjelmia pitää mitenkään muuttaa.
MySQL  tallettaa  rividatan  ja  indeksien  datan  eri  paikkaan. Indeksointi  ulottuu siis kaikkiin  järjestelmän  tukemiin tietotyyppeihin.
Useimmat MySQL:n indeksityypit (perusavain (PRIMARY KEY), yksikäsitteinen  indeksi(UNIQUE), B-puu-indeksi(INDEX)ja teksti-indeksi (FULLTEXT))   tallennetaan   B-puu-tietorakenteeseen.   
Pisteisiin,   viivoihin   ja monikulmioihin    (spatial)    liittyvät    tietotyypit    tallennetaan    R-puihin,ja MEMORY-taulut  tukevat  myös hajautusindeksejä  (hashindex).

## Hajautusindeksit
Hajautusindeksit   voivat   nopeuttaa   kyselyitä   erityisesti   sellaisten   taulujen kohdalla,    jotka    mahtuvat    melkein    kokonaan    keskusmuistiin. InnoDB–tallennusmoottorin   mekanismi   valvoo   taululle   määritettyihin   indekseihin kohdistuvia  indeksihakuja.  Jos  mekanismi  huomaa,  että  kyselyt  hyötyisivät hajautusindeksistä, tallennusmoottori luo sellaisen automaattisesti.Hajautusindeksi  rakentuu aina  perustuen  taulussa  jo  olevaan B-puu-indeksiin.
 
### Yksilöivät indeksit 
Useimmat DBMS:t muodostavat yksilöivän indeksin (UNIQUE INDEX) automaattisesti perusavaimen yhteyteen. On mahdollista luoda indeksi myös viiteavaimille nopeuttamaan liitosten muodostamista. 
Indeksi on ikään kuin järjestetty kirjan hakemisto josta on suorat viittaukset oikealle kohdalle dataan. 
Kuten relaatioteorian periaatteisiin kuuluu, taulun rivit ovat normaalisti epäjärjestyksessä. 
 
### Ryvästävät indeksit 
Ryvästävä indeksi (CLUSTERED INDEX, joissakin DBMS:ssä) on indeksi, joka järjestää myös taulun rivit indeksin mukaan. Olemassa voi olla vain yksi ryvästävä indeksi yhtä taulua kohti. 
 
### Indeksien luonti 
CREATE INDEX –komennolla luodaan indeksi, jolle voidaan antaa koko tietokannan sisällä pätevä yksilöllinen nimi. 
Jos käytetään UNIQUE-määritettä CREATE-sanan perässä, voidaan luoda yksilöivä indeksi. 
 
Esimerkiksi indeksin luominen Tyontekija-taulun sukunimi-kentälle (oletuksena usein haettu hakuperuste). 
``` 
CREATE INDEX Tyontekija_sukunimi ON Tyontekija(sukunimi); 
``` 
Yksilöivä indeksi hetu-kentälle 
``` 
CREATE UNIQUE INDEX Tyontekija_hetu ON Tyontekija(hetu); 
```
### Indeksien poisto 
Indeksin poisto tapahtuu DROP INDEX-komennolla: 
```
DROP INDEX Tyontekija_sukunimi; 
DROP INDEX Tyontekija_hetu; 
```
### INDEKSIEN KÄYTTÖ
**Indeksoinnin etuja:**
hakujen, liitosten, ryhmittelyjen ja lajittelujen nopeutuminen–pakottaa yksilöimään rivit

**Kannattaa indeksoida:**
* perusavaimet ja viiteavaimet (jokainen viiteavain erikseen!)
* sarakkeet, joita haetaan tai järjestetään usein (ja joissa on paljontietoa eli rivejä on tuhansia tai miljoonia)

Kannattaa huomioida, että indeksointi vie levytilaa ja aikaa luotaessa ja erityisesti silloin, kun on tehtävä muutoksia tietokantarakenteeseen.

**Ei kannata indeksoida:**
* harvoin haussa käytettäviä sarakkeita
* sarakkeita, joiden selektiivisyys on huono
*
**PEUKALOSÄÄNTÖ INDEKSIN LUOMISELLE (NK. 3 TÄHDENINDEKSI)**

1.   Ota kaikki WHERE-kohdassa mainitut sarakkeet
2.   Lisää ORDER BY –kohdassa mainitut sarakkeet
3.   Laita loppuun SELECT-lauseessa olevat sarakenimet.

Esimerkki: Sopiva indeksi kyselylle.
```
SELECT a, b, c FROM t WHERE d= ’x’ ORDER BY b 
```
olisi 
```
CREATE INDEX i_dbac ON t(d,b,a,c);
```
Jos kaikki SELECT:ssä mainitut sarakkeet on mukana indeksissä, kyse on paksusta indeksistä.

## Merkistö eli Character set ja Collation-asetukset

Käytössä oleva merkistö voidaan määritellä koko järjestelmän tasolla, tietokantakohtaisesti tai taulukohtaisesti. Lyhyesti selitettynä merkistö määrittää, mitä kirjaimia on käytössä. Esimerkiksi englannin kielessä ei ole ääkkösiä, Kiinassa ja Venäjällä on kokonaan eri merkit käytössä kuin länsimaissa jne. http://dev.mysql.com/doc/refman/5.1/en/charset-mysql.html

MySQL:ssä oletusmerkistö on latin1_swedish historiallisista syistä. Joskus on syytä vaihtaa merkistö utf8:ksi, eli siis me teemme yleensä näin puhtaasti suomalaisissa tietokannoissa.

Käytössä olevat merkistöt löytyvät komennolla
**show character set; **

Collation-määre kertoo järjestelmälle, miten kirjaimet lajitellaan. Esimerkiksi skandinaavisissa aakkosissa a ja ä ovat aakkosjärjestyksessä aivan eri ”päissä”. Monissa vaikkapa tsekinkielessä tai espanjassa on erilaisia ”vänkkyröitä” kirjainten päällä ja monissa kielissä on useita ”s” kirjainta merkitseviä kirjaimia. Collation kertoo, miten ne lajitellaan.  http://stackoverflow.com/questions/4538732/what-does-collation-mean

Esimerkiksi kaikki utf8-merkistöön liittyvät kollatiot voi hakea esimerkiksi komennolla
```
show collation like 'utf8%';
```
Kollation lopussa olevat kaksi viimeistä kirjainta kertovat, käsitelläänkö isot ja pienet kirjaimet erillisinä vai samoina.
    • _ci on ”case insensitive”
    • _cs on ”case sensitive”.
    • _bin on ”binary”, merkkejä lajitellaan binäärilukuina.

Oletuksena MySQL ei erottele pieniä ja isoja kirjaimia. On olemassa vain muutama kollatio, joka tukee erikseen isoja ja pieniä kirjaimia. 

Ne voidaan selvittää lauseella:
```
MariaDB [(none)]> show collation like '%cs';
+--------------------+---------+----+---------+----------+---------+
| Collation          | Charset | Id | Default | Compiled | Sortlen |
+--------------------+---------+----+---------+----------+---------+
| latin1_general_cs  | latin1  | 49 |         | Yes      |       1 |
| latin2_czech_cs    | latin2  |  2 |         | Yes      |       4 |
```
Skandit ja muut erikoismerkit aiheuttavat loppumattomasti päänvaivaa taulujen ja tietokantojen nimissä, vaikka käytössä oleva merkistö niitä tukisikin. Siksi niiden käyttöä kannattaa välttää, ellei siihen ole erityistä syytä.

### Jatketaan seuraavaksi lemmikkieläintietokannan tutkimista.
Käytetty esimerkki sotii tietokannan suunnittelun yhtä olennaisinta perussääntöä vastaan. Nimittäin samassa taulussa esiintyy omistajatieto useampaan kertaan.

Olisi varmaankin järkevää poistaa omistajatieto pet-taulusta ja luoda uusi taulu, owner, joka pitää kirjaa omistajista:
```
use test_db;

CREATE TABLE owner (
first_name varchar(255),
last_name varchar(255),
address varchar(255),
id int PRIMARY KEY AUTO_INCREMENT
);
```


**id** on nyt määritelty pääavaimeksi. Se ei voi jäädä tyhjäksi (Null=NO). Auto_increment:in avulla avaimen arvo kasvaa automaagisesti aina kun lisätään uusi tietue. Oikeastaan kaikki kentät olisi voinut määritellä NOT NULL:iksi, mutta olkoon nyt näin.

Avataan MYSQL Workbench ja luodaan ns. scripteistä ER-malli. Muokataan ER-mallia, jotta saadaan hiukan älykkäämpi relaatiomalli.


![](images/08_01_pet_owner_er.png)

Lisätään Owner -tauluun omistajia. Ota reposta INSERT_into.sql tiedosto ja kopioi sieltä Owner taulun henkilöt mysql komentoriville.

Kun edellä pet-tauluun lisättiin uusi kenttä ownerID, soluille ei annettu arvoja. Toisin sanoen missään ei olla vielä määritelty, kuka omistaa minkäkin elukan. Se voidaan tehdä UPDATE- lauseella. Fluffyn omistaja on Harold. Haroldin ownerID on 1.

Luodaan ensin pet -tauluun ownerId -kenttä:
```
MariaDB [test_db]> ALTER TABLE pet ADD COLUMN ownerID INT;
Query OK, 0 rows affected (0.037 sec)
Records: 0  Duplicates: 0  Warnings: 0
```
Lisätään sitten riippuvuus kahden taulun välille:
```
MariaDB [test_db]> ALTER TABLE pet ADD CONSTRAINT FOREIGN KEY(ownerID) REFERENCES owner(id);
Query OK, 8 rows affected (0.131 sec)              
Records: 8  Duplicates: 0  Warnings: 0
```
Lisätään vielä riippuvuudet:

Ota Moodlesta update pet scriptit ja kopio sql editoriin. Update pet löytyy INSERT_into.sql file:n lopusta.
```
-Update pet
UPDATE pet SET ownerID=1 WHERE name='Fluffy';
UPDATE pet SET ownerID=2 WHERE name='Claws'; 
UPDATE pet SET ownerID=1 WHERE name='Buffy'; 
UPDATE pet SET ownerID=3 WHERE name='Fang'; 
UPDATE pet SET ownerID=4 WHERE name='Bowser';
```
Useammankin arvon voi antaa kerralla.
```
UPDATE pet SET ownerID=3 WHERE name='Chirpy' OR name='Whistler'; 
```
### Harjoitus: Katsotaan mitä kuuluu snake elukalle.

Lisää puuttuva ownerID: Ratkaisu alla... koeta ensin itse miettiä kuinka saat haettua owner id:n

Tässä vaihtoehtoinen tapa, jossa ownerID haetaan nimen perusteella. Ratkaisu vaatii kahden sisäkkäisen SELECT-lauseen käytön.
```
MariaDB [test_db]> UPDATE pet SET ownerID=(SELECT ownerID FROM owner 
WHERE first_name='Benny') WHERE name='Fang' OR name='Slim';
Query OK, 1 row affected (0.034 sec)
Rows matched: 2  Changed: 1  Warnings: 0
```
Tsekataan lopputulos:
```
MariaDB [test_db]> select * from pet;
+----------+--------+---------+------+------------+------------+----+---------+
| name     | owner  | species | sex  | birth      | death      | id | ownerID |
+----------+--------+---------+------+------------+------------+----+---------+
| Fluffy   | Harold | cat     | f    | 1993-02-04 | NULL       |  1 |       1 |
| Claws    | Gwen   | cat     | m    | 1994-03-17 | NULL       |  2 |       2 |
| Buffy    | Harold | dog     | f    | 1989-05-13 | NULL       |  3 |       1 |
| Fang     | Benny  | dog     | m    | 1990-08-27 | NULL       |  4 |       3 |
| Bowser   | Diane  | dog     | m    | 1989-08-31 | 1995-07-29 |  5 |       4 |
| Whistler | Gwen   | bird    |      | 1997-12-09 | NULL       |  6 |       3 |
| Chirpy   | Gwen   | bird    | f    | 1998-09-11 | NULL       |  7 |       3 |
| Slim     | Benny  | snake   | m    | 1996-04-29 | NULL       |  8 |       3 |
+----------+--------+---------+------+------------+------------+----+---------+
8 rows in set (0.001 sec)
```

Yritetäänpä tuhota omistaja.
```
MariaDB [test_db]> DELETE from owner WHERE first_name='Diane';
ERROR 1451 (23000): Cannot delete or update a parent row: a foreign key constraint fails (`test_db`.`pet`, CONSTRAINT `pet_ibfk_1` FOREIGN KEY (`ownerID`) 
REFERENCES `owner` (`ownerID`))
MariaDB [test_db]> 
```
Toisin sanoen siitä taulusta, joka on ”yhden suhde moneen”-relaatiossa ”yksi” eli isäntätaulu, ei voi tuhota tietuetta, ellei poistoa ole erikseen sallittu lapsitaulun ON DELETE CASCADE- määrityksellä .



### Tapahtumien eli transaktioiden hallinta
Tapahtuma on yhden tai useamman käskyn kokonaisuus, joka suoritetaan joko kokonaan tai ei ollenkaan.

BEGIN aloittaa tapahtuman (MySQL, monissa tuotteissa **BEGIN  TRANSACTION**)
**COMMIT** hyväksyy tapahtuman
**ROLLBACK** peruuttaa tapahtuman

**Perinteinen tilisiirtoesimerkki**

![](images/08_02_perinteinen_tilisiirto01.png){width=50% height=50%}

![](images/08_02_perinteinen_tilisiirto02.png){width=50% height=50%}

### Transaktio sovelluksessa

Transaktion alku on joko:
* eksplisiittinen: komento begin transaction
* implisiittinen: ei komentoa, vaan             ohjelman(osan) käynnistys tai edellisen transaktion päättyminen.

Transaktio päättyy joko:
* vahvistukseen: komento commit
* peruutukseen: komentoon rollback
* DBMSn suorittamaan automaattiseen peruutukseen, jos ei vahvistusta tule


### Autocommit usein oletusarvona

Mm. ODBC, JDBC ja MySQL ml MariDB käyttävät autocommit -tilamuuttujaa, jonka arvot ovat ON tai OFF. Oletusarvo on ON.

Autocommit ei sovi kaikkiin sovelluksiin.
Yleensä on annettava aluksi komento set autocommit=off tai set autocommit=0

**MariaDB**
```
MariaDB [(none)]> select @@autocommit;  
+--------------+  
| @@autocommit |  
+--------------+  
|            1 |  
+--------------+  
1 row in set (0.000 sec)  
```
```
MariaDB [(none)]> SET AUTOCOMMIT=0;
Query OK, 0 rows affected (0.000 sec)

MariaDB [(none)]> select @@autocommit;
+--------------+
| @@autocommit |
+--------------+
|            0 |
+--------------+
1 row in set (0.000 sec)

MariaDB [(none)]> SET AUTOCOMMIT=on;
Query OK, 0 rows affected (0.000 sec)
```

### Atomisuus
Transaktio suoritetaan aina jakamattomana eli atomisena (atomic)
***KAIKKI TAI EI MITÄÄN!***


SQL- Eristyvyystasot / samanaikaisuuden ongelmat

Eristyvyystaso ilmaisee transaktion eristyvyyttä eli riippumattomuutta muista samanaikaisista transaktioista.
Tavoitteena ehkäistä samanaikaisuuden ongelmia.

### Ongelmatilanteet:

**Hukatun päivityksen ongelma (lost update):** 
a) päällekirjoitus.

**Keskeneräisen käsittelyn ongelma (dirty read)**
a) luetaan peruutettava
b) päivitetään peruutettava

**Ristiriitaisen tulkinnan ongelma (inconsistent analysis)**
a) eritahtinen päivitys (nonrepeatable read)
b) uusien rivien ongelma (phantom rows)

**Hukatun päivityksen ongelma**

AINA ESTETTÄVÄ
Henkilö T1 ja T2 käyttävät samaa tiliä
Tilille pitäisi jäädä 100eur, mutta sinne jää 300eur
![](images/08_03_hukattu_paivitys.png){width=50% height=50%}

### Keskeneräisen luku

**"DIRTY READ"**
Transaktiota ei vielä komittoitu
T1 luki tilin saldon keskeneräisestä tilanteesta, joka ei jäänyt voimaan.
![](images/08_04_keskenerainen_luku.png){width=50% height=50%}

### Keskeneräisen päivitys
**AINA ESTETTÄVÄ**
T1 lukee tilin saldon keskeneräisestä tilanteesta, jossa transaktioita T2 ei ole vielä hyväksytty.
![](images/08_05_keskenerainen_paivitys.png){width=50% height=50%}

### Eritahtinen päivitys

**Nonrepeatable read**
Transaktio T1 luki tilin saldoksi 4300eur (600+1000+2700), vaikka se oli viimeisen tilin lukuhetkellä 4600eur.
![](images/08_06_eritahtinen_paivitys.png){width=50% height=50%}

### Uusien rivien ongelma

**Phantom rows**
Transaktio T1 luki 4600eur, vaikka viimeisen tilin lukuhetkellä tilien yhteissumma oli 5100eur.

![](images/08_07_uusien_rivien_ongelma.png){width=50% height=50%}

## Eristyvyystasojen asettaminen

Samanaikaisuuden ongelmien käsittelemiseksi tietokantatuotteissa voidaan asettaa haluttu eristyvyystaso.
**Eristyvyystasot ovat:**
![](images/08_08_eriytyvyys_tasot.png){width=50% height=50%}

Asettaminen käytännössä MariaDB:ssä:
```
MariaDB [(none)]> **use devices;**
Database changed
MariaDB [devices]> S**ELECT @@tx_isolation;**
+-----------------+
| @@tx_isolation  |
+-----------------+
| REPEATABLE-READ |
+-----------------+
1 row in set (0.000 sec)

MariaDB [devices]> **SET tx_isolation = 'READ-UNCOMMITTED';**
Query OK, 0 rows affected (0.000 sec)

MariaDB [devices]> **SET tx_isolation = 'READ-COMMITTED';**
Query OK, 0 rows affected (0.000 sec)

MariaDB [devices]> **SET tx_isolation = 'REPEATABLE-READ';**
Query OK, 0 rows affected (0.000 sec)

MariaDB [devices]> **SET tx_isolation = 'SERIALIZABLE';**
Query OK, 0 rows affected (0.000 sec)
```

### Testataan toimivuudet

Luodaan ensin tietokanta, johon voidaan kohdentaa transaktioita esimerkiksi **kahdesta eri terminaalista.**

#### Esimerkki 1. 
Luodaan ensin tietokanta:
```
MariaDB [devices]> create database transaktio;
Query OK, 1 row affected (0.000 sec)
```
Vaihdetaan käytettävä tietokanta
```
MariaDB [(none)]> use transaktio;
Database changed
```
Luodaan tietokantaa taulu:
```
MariaDB [transaktio]> CREATE TABLE t (f INT) ENGINE=InnoDB;
Query OK, 0 rows affected (0.058 sec)
```
Aloitetaan transaktio:
```
MariaDB [transaktio]> BEGIN;
Query OK, 0 rows affected (0.000 sec)
```
Lisätään dataa tauluun:
```
MariaDB [transaktio]> INSERT INTO t(f) VALUES (1);
Query OK, 1 row affected (0.000 sec)
```
Katsotaan mitä tapahtui:
```
MariaDB [transaktio]> SELECT * FROM t;
+------+
| f    |
+------+
|    1 |
+------+
1 row in set (0.000 sec)
```
Ajetaan palautus:
```
MariaDB [transaktio]> ROLLBACK;
Query OK, 0 rows affected (0.034 sec)
```
Tarkistetaan tilanne:
```
MariaDB [transaktio]> SELECT * FROM t;
Empty set (0.000 sec)
```
#### Esimerkki 2
**Yhteys 1:**

```
MariaDB [transaktio]> drop table t;
Query OK, 0 rows affected (0.017 sec)

MariaDB [transaktio]> CREATE TABLE t (f INT) ENGINE=InnoDB;
Query OK, 0 rows affected (0.055 sec)

MariaDB [transaktio]> INSERT INTO t(f) VALUES (1);
Query OK, 1 row affected (0.034 sec)

MariaDB [transaktio]> SELECT * FROM t;
+------+
| f    |
+------+
|    1 |
+------+
1 row in set (0.000 sec)
```
**Yhteys 2:**
```
MariaDB [transaktio]> SELECT * FROM t;
+------+
| f    |
+------+
|    1 |
+------+
1 row in set (0.000 sec)
```
Yhteys 1:
```
MariaDB [transaktio]> SET AUTOCOMMIT=0;
Query OK, 0 rows affected (0.000 sec)

MariaDB [transaktio]> INSERT INTO t(f) VALUES (2);
Query OK, 1 row affected (0.000 sec)
```
Yhteys 2:
```
MariaDB [transaktio]> SELECT * FROM t;
+------+
| f    |
+------+
|    1 |
+------+
1 row in set (0.000 sec)
```
Yhteys 1:
```
MariaDB [transaktio]> COMMIT;
Query OK, 0 rows affected (0.034 sec)
```
Yhteys 2:
```
MariaDB [transaktio]> SELECT * FROM t;
+------+
| f    |
+------+
|    1 |
|    2 |
+------+
2 rows in set (0.000 sec)
```
#### Lukulukot päivityksessä
**Esimerkki 3**

Halutaan lisätä MAX-arvoa pykälän suurempi luku
kahdella eri yhteydellä:
Yhteys 1:
```
MariaDB [transaktio]> BEGIN;
Query OK, 0 rows affected (0.000 sec)

MariaDB [transaktio]>   SELECT MAX(f) FROM t;
+--------+
| MAX(f) |
+--------+
|      2 |
+--------+
1 row in set (0.031 sec)
```
Yhteys 2:
```
MariaDB [transaktio]>   SELECT MAX(f) FROM t;
+--------+
| MAX(f) |
+--------+
|      3 |
+--------+
1 row in set (0.000 sec)
```
Yhteys 1:
```

MariaDB [transaktio]> SET AUTOCOMMIT=0;
Query OK, 0 rows affected (0.000 sec)

MariaDB [transaktio]>  INSERT INTO t(f) VALUES (4);
Query OK, 1 row affected (0.001 sec)
```
Yhteys 2:
```
MariaDB [transaktio]>   SELECT MAX(f) FROM t;
+--------+
| MAX(f) |
+--------+
|      3 |
+--------+
1 row in set (0.000 sec)
```
Yhteys 1:
```
MariaDB [transaktio]> SELECT * FROM t;
+------+
| f    |
+------+
|    1 |
|    2 |
|    3 |
|    4 |
+------+
4 rows in set (0.000 sec)
```
Yhteys 2:
```
MariaDB [transaktio]>   SELECT MAX(f) FROM t;
+--------+
| MAX(f) |
------+
|      3 |
+--------+
1 row in set (0.000 sec)

MariaDB [transaktio]> INSERT INTO t(f) VALUES(4);
Query OK, 1 row affected (0.000 sec)

MariaDB [transaktio]> COMMIT;
Query OK, 0 rows affected (0.035 sec)
```
Yhteys 1:
```
MariaDB [transaktio]> COMMIT;
Query OK, 0 rows affected (0.033 sec)

MariaDB [transaktio]> SELECT * FROM t;
+------+
| f    |
+------+
|    1 |
|    2 |
|    3 |
|    4 |
|    4 |
+------+
5 rows in set (0.000 sec)
```
Haluttiin luvut 4 ja 5, mutta on kaksi nelosta

**Esimerkki 4:**
Yhteys 1:
```
MariaDB [transaktio]>  DELETE FROM t WHERE f=4;
Query OK, 2 rows affected (0.031 sec)

MariaDB [transaktio]>   SELECT * FROM t;
+------+
| f    |
+------+
|    1 |
|    2 |
|    3 |
+------+
3 rows in set (0.000 sec)

MariaDB [transaktio]> BEGIN;
Query OK, 0 rows affected (0.036 sec)

MariaDB [transaktio]> SELECT MAX(f) FROM t FOR UPDATE;
+--------+
| MAX(f) |
+--------+
|      3 |
+--------+
1 row in set (0.000 sec)

MariaDB [transaktio]> INSERT INTO t(f) VALUES (4);
Query OK, 1 row affected (0.000 sec)

```
Yhteys 2:
```

MariaDB [transaktio]> SELECT MAX(f) FROM t FOR UPDATE;

```
Ei tapahtu mitään, koska Yhteys 1 lukinnut päivitystä varten tapahtuman.

Yhteys 1:
```
MariaDB [transaktio]> COMMIT;
Query OK, 0 rows affected (0.034 sec)
```
Yhteys 2:
```
ERROR 1205 (HY000): Lock wait timeout exceeded; try restarting transaction
MariaDB [transaktio]> SELECT MAX(f) FROM t FOR UPDATE;
+--------+
| MAX(f) |
+--------+
|      4 |
+--------+
1 row in set (0.000 sec)
```
## Samanaikaisuuden hallinnan toteutustekniikoita
DBMSien tarjoamat toteutustavat:

* lukitus
* versiointi


**Tekniikkana lukitus**

DBMS voi hallita samanaikaisuutta esim. lukitsemalla tietoja (nk. pessimistinen menetelmä).

* Lukulukko on jaettu (S shared), muut saavat lukea,
* Kirjoituslukko yksityinen (X exclusive), poissulkeva.
Käytännössä vanhin ja yleisin tekniikka.

**Tekniikkana versiointi**

Jokainen transaktio käsittelee omaa versiotaan (eli omaa kopiota) datasta.
Vasta commit-hetkellä DBMS ilmoittaa mahdollisen konfliktin: että joku muu on ehtinyt muuttaa "originaali" dataa.

Nopein voittakoon, muut aloittakoot alusta!
Kyky uudelleenaloitukseen konfliktin jälkeen on toteutettava ohjelman rakenteeseen.
Versiointi on optimistinen menetelmä "nyt ei voi tulla konfliktia", lukinta pessimistinen "kumminkin nyt tulee konflikti"

# MySQL - Osa 7 - Näkymät - VIEWS

Sisältö:

- Näkymät ja näkymien käyttö
- Indeksit - Suorituskyvyn optimointi


SQL-kyselyistä tulee helposti hyvin pitkiä, jos joudutaan hakemaan tietoa monesta taulusta. Tätä tuskaa helpottamaan on käytössä Näkymät (Views) ja Tallennetut Aliohjelmat (Stored Procedures).
Trigger eli ”Liipaisin” käynnistää tietokannassa jonkin ennalta määritellyn toiminnon, kun kentän sisältö muuttuu.

## View eli näkymä

Näkymä sisältää rivejä ja sarakkeita, aivan kuten oikea taulu. 
Näkymän kentät on koottu oikeiden taulujen kentistä määriteltyjen hakuehtojen mukaisesti. 
Näkymään voidaan kohdistaa SQL-lauseita, ja tulos esitetään samalla lailla kuin tulos tulisi yhdestä taulusta.

* Näkymä (view) on looginen näyte tietokannan tauluista tai näkymistä
* Näkymä ei voi sisältää SELECT INTO, ORDER BY tai COMPUTE –ehtoja (MariaDB:ssä näin)
* Ei voi viitata väliaikaisiin tauluihin
* Tarvitaan SELECT-oikeus tauluihin

### Näkymän luonti
```
MariaDB [sql_kyselyt]> CREATE VIEW Kulut AS SELECT Salary, Department_ID FROM employees;
Query OK, 0 rows affected (0.036 sec)

MariaDB [sql_kyselyt]> SELECT * FROM Kulut;
+----------+---------------+
| Salary   | Department_ID |
+----------+---------------+
| 24000.00 |            90 |
| 17000.00 |            90 |
| 17000.00 |            90 |
|  9000.00 |            60 |
|  6000.00 |            60 |
|  4800.00 |            60 |
|  4800.00 |            60 |
|  4200.00 |            60 |
| 12000.00 |           100 |
|  9000.00 |           100 |
|  8200.00 |           100 |
|  7700.00 |           100 |
|  7800.00 |           100 |
|  6900.00 |           100 |
```

### Näkymän käyttö
* Näkymän kautta voidaan päivittää tai lisätä tietoja, mutta yleensä vain yhteen tauluun.
* Näkymässä ei saa olla DISTINCT-määrettä, HAVING-lausetta eikä alikyselyä
* 
* Näkymä poistetaan DROP VIEW -komennolla
* Jos näkymässä on laskettuja sarakkeita, view-määrittelyn nimen jäljessä täytyy luetella sarakkeiden nimet:

Esimerkki - Lasketaan uudet palkat korotuksen jälkeen:
```
MariaDB [sql_kyselyt]> CREATE VIEW Palkankorotus (First_Name, Last_Name, UusiPalkka) AS SELECT First_Name, Last_Name, SALARY*1.05 FROM employees;
Query OK, 0 rows affected (0.036 sec)
```
```
MariaDB [sql_kyselyt]> select * from Palkankorotus;
+-------------+-------------+------------+
| First_Name  | Last_Name   | UusiPalkka |
+-------------+-------------+------------+
| Steven      | King        | 25200.0000 |
| Neena       | Kochhar     | 17850.0000 |
| Lex         | De Haan     | 17850.0000 |
| Alexander   | Hunold      |  9450.0000 |
| Bruce       | Ernst       |  6300.0000 |
| David       | Austin      |  5040.0000 |

```
Vinkkejä näkymien luontiin

* Testaa ensin kysely ja luo vasta sitten näkymä
* Jos taulun nimi muuttuu, luo kysely jossa on alkuperäisen taulun kentät; tällä tavoin sinun ei tarvitse muuttaa olemassa olevia SELECT- ym. käskyjä
* Näkymillä voit suojata tietoja käyttäjäryhmittäin
* Voit rakentaa näkymän toisen näkymän päälle, mutta se voi vaikeuttaa ymmärtämistä
* Joissakin tuotteissa nk. Materialized view (eli indexed view), jossa näkymä tallennettu levylle (ikään kuin tauluksi)
* Näkymistä lisätietoa täältä: http://www.mysqltutorial.org/mysql-views-tutorial.aspx

HUOM: MySQL Workbenchissä klikkaa Views-kohtaa hiiren oikealla napilla ja sitten Create View.
Erillinen editori aukeaa, johon koodin voi kirjoittaa.

**Kokeillaan muutamia esimerkkejä näkymistä:**

Luodaan Opiskelijat kantaan näkymä Suoritukset.

```
MariaDB [Opiskelijat]> CREATE VIEW Suoritukset AS SELECT Sukunimi, Etunimi, Nimi, Arvosana FROM Opiskelija JOIN Tenttii ON Opiskelija.Sotu=Tenttii.Opiskelija JOIN Kurssi ON Kurssi.Kurssikoodi=Tenttii.Kurssi;     
Query OK, 0 rows affected (0.036 sec)

MariaDB [Opiskelijat]> desc Suoritukset;
+----------+-------------+------+-----+---------+-------+
| Field    | Type        | Null | Key | Default | Extra |
+----------+-------------+------+-----+---------+-------+
| Sukunimi | varchar(64) | NO   |     | NULL    |       |
| Etunimi  | varchar(32) | NO   |     | NULL    |       |
| Nimi     | varchar(64) | NO   |     | NULL    |       |
| Arvosana | double      | NO   |     | 1       |       |
+----------+-------------+------+-----+---------+-------+
4 rows in set (0.002 sec)

```
Toinen esimerkki. Halutaan tietää kunkin opiskelijan arvosanojen keskiarvo. Hyödynnetään
GROUP BY-lausetta, joka luokittelee tulostuksen valitun kentän mukaan:
```
MariaDB [Opiskelijat]> SELECT Sukunimi, Etunimi, AVG(Arvosana) FROM Suoritukset GROUP BY Sukunimi;
+------------+---------+--------------------+
| Sukunimi   | Etunimi | AVG(Arvosana)      |
+------------+---------+--------------------+
| Huusko     | Eero    |                1.6 |
| Karppinen  | Markku  |                1.8 |
| Kasvattaja | Kimmo   | 1.9166666666666667 |
| Kielinero  | Kaisa   |              2.125 |
| Kirjailija | Keijo   |                1.5 |
| Tieteilija | Taina   |                1.7 |
+------------+---------+--------------------+
6 rows in set (0.032 sec)
```
Ja lopuksi keskiarvot opintojaksoittain:
```
MariaDB [Opiskelijat]> SELECT Nimi, AVG(Arvosana) FROM Suoritukset GROUP BY Nimi;
+--------------------------------+--------------------+
| Nimi                           | AVG(Arvosana)      |
+--------------------------------+--------------------+
| Johdatus ohjelmointiin         |                  1 |
| Mikrotietokoneiden ohjelmistot |  2.107142857142857 |
| Ohjelmointi++                  |             2.0625 |
| Tietokannat                    | 2.3055555555555554 |
| Tietotekniikan propedeuttinen  | 1.5833333333333333 |
| Tietoverkot tyvvvvvvalineen    |              1.375 |
+--------------------------------+--------------------+
6 rows in set (0.001 sec)
```
# Stored procedure eli tallennettu aliohjelma
Stored procedure on tietokantaan tallennettu aliohjelma.

Aliohjelman avulla voidaan automatisoida usein toistuvia tehtäviä. 
Aliohjelmaan on mahdollista määritellä muuttujia, ehtoja ja silmukoita ja niitä voivat käyttää ulkopuoliset ohjelmat (Java, C#,PHP jne). Aliohjelmat lisäävät tietoturvallisuutta siinä mielessä, että on mahdollista määritellä, että ulkoiset sovellukset saavat käyttää tietokantaa vain aliohjelmien kautta.


## Tallennetun aliohjelman kirjoittaminen

Aliohjelman voi kirjoittaa Workbenchillä klikkaamalla hiiren oikealla napilla Stored Procedures- kohtaa.

Klassinen Hello World MySQL:n aliohjelmana:
```
DELIMITER //
CREATE PROCEDURE `p2` ()
COMMENT 'A procedure'
BEGIN
SELECT 'Hello World !';
END//
Aliohjelman kutsu tapahtuu CALL-käskyllä. Huomaa sulut
CALL p2();
Lista kaikista alio-ohjelmista
SHOW PROCEDURE STATUS;

```
Aliohjelman muuttamiseen ei ole yleispätevää työkalua. On olemassa komento ALTER PROCEDURE, mutta sen avulla voi muuttaa vain joitakin ominaisuuksia. 
Helpommalla pääsee, jos ottaa koodista kopion leikepöydälle, poistaa aliohjelman.

**DROP PROCEDURE IF EXISTS nimi;**

ja luo uuden aliohjelman ja editoi sitä Workbenchillä. Kun painetaan oikeassa alanurkassa olevaa Apply-painiketta, aliohjelma siirretään palvelimelle.
```
DELIMITER //

CREATE procedure OpiskelijaLista (OUT Sukunimi VARCHAR(32) )
DETERMINISTIC

BEGIN 
	SELECT * FROM Opiskelija;
END; //
DELIMITER ;

```

Esimerkki Tallennetun proseduurin käytöstä:
```
MariaDB [Opiskelijat]> Call OpiskelijaLista (@Sukunimi);
+-------------+---------------------+---------+-------------+-------------+-------------+--------------+--------+
| Sotu        | Sahkopostiosoite    | Etunimi | Sukunimi    | Lahiosoite  | Postinumero | Aloitusvuosi | Laitos |
+-------------+---------------------+---------+-------------+-------------+-------------+--------------+--------+
| 010101-011P | bar@foo.invalid     | Taina   | Tieteilija  | TNT 9 Y 124 | 40740       |         1996 |      2 |
| 111111-111P | eero.huusko@kamk.fi | Eero    | Huusko      | TNT 9 a 666 | 40740       |         1995 |      1 |
| 111112-111P | barfoo@bar.invalid  | Kaisa   | Kielinero   | Keskusta    | 40700       |         1995 |      3 |
```

Aliohjelmalle voi antaa syötteenä parametreja:
- CREATE PROCEDURE procName () : Parameter list is empty
- CREATE PROCEDURE procName (IN varname DATATYPE) : One input parameter. The word IN is optional because parameters are IN (input) by default.
- CREATE PROCEDURE procName (OUT varname DATATYPE) : One output parameter.
- CREATE PROCEDURE procName (INOUT varname DATATYPE) : One parameter which is both input and output.

Myös useamman parametrin käyttö on sallittu.

Seuraava demonstroidaan yhden OUT-parametrin käyttöä. 

Parametrilistassa on määritelty yksi OUT, eli output-parametri, nimeltään Lkm ja määritelty se INT-tyyppiseksi.

```
DELIMITER //

CREATE procedure OpiskelijaLkm (OUT LKM INT )
DETERMINISTIC

BEGIN 
	SELECT COUNT(*) INTO LKM FROM Opiskelija;
END; //
DELIMITER ;
```

SELECT INTO-lauseessa lasketaan, montako opiskelijaa Opiskelija-taulussa on.  
Tallennetaan tulos muuttujaan Lkm. Muuttujan arvo palautuu kutsuvalle ohjelmalle, joka tässä tapauksessa on MariaDB:n komentorivi.

```
MariaDB [Opiskelijat]> CALL OpiskelijaLkm(@N);
Query OK, 1 row affected (0.001 sec)
```
Muuttujan arvo saadaan kutsuvassa ohjelmassa selville
```
MariaDB [Opiskelijat]> SELECT @N;
+------+
| @N   |
+------+
|    9 |
+------+
1 row in set (0.000 sec)
```

Aliohjelmia voi käyttää myös oikeasti hyödyksi. Monesti on esimerkiksi vaikea muistaa jonkin MySQL-lauseen syntaksia. Sen voi kirjoittaa aliohjelmaksi. Seuraava esimerkki lisää uuden kurssin Opintojakso-tauluun. Se ottaa syötteeksi kurssin nimen, laajuuden ja koodin IN-tyyppisinä parametreina ja lisää ne tauluun INSERT:llä.
```
DELIMITER //
DROP procedure IF EXISTS `Raute_Tietokannat`;
CREATE PROCEDURE `Raute_Tietokannat`(IN K VARCHAR(45),IN L INT, IN N VARCHAR(45))
BEGIN
INSERT INTO Kurssi VALUES('',K,L,N);
END; //
DELIMITER ;
```

# Linkit

[07: Vaativammat tietokantakyselyt](/07_vaativammat_tietokantakyselyt) <----------> [09: Indeksit ja Tietokannan optimointi](/09_indeksit_ja_tietokannan_optimointi)
